import logging
from datetime import datetime

from opencensus.stats import stats
from opencensus.tags import tag_map, tag_key, tag_value

from openldap_opencensus_stats.ldap_sync_statistic import LdapSyncStatistic


class SyncMetricSet:
    """
    Set of statistics for LDAP server synchronization.  This set
    represents a LDAP tree base, and contains statistics for
    the offset of each server that services that tree.

    So, if the tree is _dn=example,dn=org_ and the tree is serviced
    by servers ldap1, ldap2, and ldap3, then the resultant stats
    will:
    * Have a name of 'sync/{servername}/offset'
    * Have a BaseDN tag of 'dn=example,dn=org'
    * Have floating point values representing the number of seconds
      that the latest timestamp of their tree differs from the
      timestamp of the most recent change among the entire cluster
    """
    def __init__(self,
                 base_dn=None,
                 ldap_servers=None):
        if not base_dn:
            logging.error('INTERNAL: Sync metric set created without the base DN')
            raise ValueError('INTERNAL: Sync metric set created without the base DN')
        self._base_dn = base_dn
        if not ldap_servers:
            logging.error('INTERNAL: Sync metric set created without any LDAP servers')
            raise ValueError('INTERNAL: Sync metric set created without any LDAP servers')
        self.timestamp_attribute = 'contextCSN'

        self._statistics = {}
        for ldap_server in ldap_servers:
            self._statistics[ldap_server] = LdapSyncStatistic(
                name=f'sync/{ldap_server.database}/offset',
                description='Offset in seconds from the most recent update'
            )

    def collect(self):
        # Setup
        #################################################
        mmap = stats.stats.stats_recorder.new_measurement_map()

        # Main Processing
        #################################################
        watermarks = {}
        for ldap_server in self._statistics.keys():
            result = ldap_server.query_dn_and_attribute(
                dn=self._base_dn,
                attribute=self.timestamp_attribute
            )
            if (result and result[0]):
                result_str = result[0].decode('utf-8')
                segments = result_str.split('#')
                watermarks[ldap_server.database] = datetime.strptime(
                    segments[0],
                    '%Y%m%d%H%M%S.%fZ'
                )  # Record the timestamp

        high_water_mark = max(watermarks.values())
        for ldap_server, stat in self._statistics.items():
            if (ldap_server.database in watermarks):
                this_watermark = watermarks[ldap_server.database]
                offset = (high_water_mark - this_watermark).total_seconds()
                stat.collect(
                    ldap_server=ldap_server,
                    measurement_map=mmap,
                    offset=offset
                )
        # Record/Publish the data
        tmap = tag_map.TagMap()
        tmap.insert(
            tag_key.TagKey('BaseDN'),
            tag_value.TagValue(self._base_dn)
        )
        mmap.record(tmap)



from openldap_opencensus_stats.configuration import Configuration   # noqa: F401
from openldap_opencensus_stats.ldap_server import LdapServer        # noqa: F401
from openldap_opencensus_stats.ldap_statistic import LdapStatistic  # noqa: F401




# The transformations process in the order listed here.
from openldap_opencensus_stats.config_transformers.base \
    import ConfigurationTransformationChainSingleton
from openldap_opencensus_stats.config_transformers.snake_case \
    import SnakeCaseConfigurationTransformer                                   # noqa: F401
from openldap_opencensus_stats.config_transformers.metric_dn \
    import MetricDnConfigurationTransformer                                    # noqa: F401
from openldap_opencensus_stats.config_transformers.copy_for_each_database \
    import CopyForEachDatabaseConfigurationTransformer                         # noqa: F401
from openldap_opencensus_stats.config_transformers.child_object \
    import ChildObjectConfigurationTransformer                                 # noqa: F401
from openldap_opencensus_stats.config_transformers.metric_name_interpolation \
    import MetricNameInterpolationConfigurationTransformer                     # noqa: F401
from openldap_opencensus_stats.config_transformers.metric_name \
    import MetricNameConfigurationTransformer                                  # noqa: F401
from openldap_opencensus_stats.config_transformers.metric_query_dn \
    import MetricQueryDnConfigurationTransformer                               # noqa: F401

ConfigurationTransformationChainSingleton().register(MetricDnConfigurationTransformer)



import copy
import logging
import re

import ldap

from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class ChildObjectConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        if not isinstance(configuration, dict):
            raise ValueError('The top level of the configuration must be a mapping.')

        config = copy.deepcopy(configuration)
        for server_config in config.get('ldap_servers'):
            if not server_config.get('sync_only', False):
                ldap_server = ChildObjectConfigurationTransformer.get_ldap_server(server_config)
                server_config['object'] = ChildObjectConfigurationTransformer.process_objects_for_ldap_server(
                    configuration=server_config.get('object', {}),
                    dn='',
                    ldap_server=ldap_server
                )

        return config

    @staticmethod
    def process_objects_for_ldap_server(configuration, dn, ldap_server):
        if not configuration or not ldap_server:
            raise ValueError(
                'ChildObjectConfigurationTransformer.process_objects_for_ldap_server: All arguments must exist'
            )

        if isinstance(configuration, dict):
            config = {}
            for key, value in configuration.items():
                proper_value = ChildObjectConfigurationTransformer.process_objects_for_ldap_server(
                    value,
                    configuration.get('computed_dn', dn),
                    ldap_server
                )
                if key == 'children':
                    if not isinstance(proper_value, dict):
                        raise TypeError("A configuration entry of 'children' must contain a mapping.")

                    for child_dn in ChildObjectConfigurationTransformer.get_child_keys(dn, ldap_server):
                        child_value = copy.deepcopy(proper_value)
                        child_value['computed_dn'] = child_dn
                        child_value['computed_rdn'] = re.sub(r',.*', '', child_dn)
                        new_key = re.sub(r',.*', '', re.sub(r'^[^=]*=', '', child_dn))
                        if re.match(child_value.get('rdn'), child_value.get('computed_rdn')):
                            config[new_key] = child_value
                else:
                    config[key] = proper_value
            return config
        elif isinstance(configuration, list):
            return [
                ChildObjectConfigurationTransformer.process_objects_for_ldap_server(item, dn, ldap_server)
                for item in configuration
            ]
        else:
            return configuration

    @staticmethod
    def get_child_keys(dn, ldap_server):
        ldap_result = ldap_server.query(dn=dn, scope=ldap.SCOPE_ONELEVEL)
        if not ldap_result:
            logging.warning(f"Found no child objects for {dn} on LDAP database {ldap_server.database}!")
        return [
            result_item[0]
            for result_item in ldap_result
        ]



import logging
import re

from openldap_opencensus_stats.ldap_server import LdapServerPool


class ConfigurationTransformer:
    transformers = []

    def __init_subclass__(cls, **kwargs):
        ConfigurationTransformationChainSingleton().register(cls)

    @staticmethod
    def process(configuration):
        return configuration

    @staticmethod
    def get_ldap_server(server_config):
        args = dict([
            (re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower(), value)
            for name, value in server_config.get('connection', {}).items()
        ])
        args['database'] = server_config.get('database')
        ldap_server = LdapServerPool().get_ldap_server(**args)
        return ldap_server


class ConfigurationTransformationChainSingleton:
    transformation_chain = []

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(ConfigurationTransformationChainSingleton, cls).__new__(cls)
        return cls.instance

    def register(self, cls):
        self.transformation_chain.append(cls)
        logging.info(f"Registered transformer: {cls.__name__}")

    def transform_configuration(self, configuration):
        configuration_result = {}
        configuration_input = configuration
        for link in self.transformation_chain:
            print(f"Calling configuration chain link {link.__name__}")
            configuration_result = link.process(configuration_input)
            configuration_input = configuration_result
        return configuration_result



import copy
import logging
import logging.config

import yaml
from time import sleep

from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformationChainSingleton
from openldap_opencensus_stats.ldap_metric_set import MetricSet
from openldap_opencensus_stats.sync_metric_set import SyncMetricSet
from openldap_opencensus_stats.ldap_server import LdapServerPool
from openldap_opencensus_stats.ldap_statistic import LdapStatistic

from opencensus.stats import stats
from opencensus.ext.prometheus import stats_exporter
import opencensus.ext.stackdriver.stats_exporter

# Make up for broken code in the Prometheus exporter
import opencensus.stats.aggregation_data
opencensus.stats.aggregation_data.SumAggregationDataFloat = opencensus.stats.aggregation_data.SumAggregationData

SUPPORTED_EXPORTERS = ['Prometheus', 'Stackdriver']


class Configuration:
    def __init__(self, config_file_name):
        if config_file_name is None:
            raise ValueError("Config file name must be supplied")
        self._config_file_name = config_file_name
        self._configuration_dict = {}
        self._sleep_time = 5
        self._metric_sets = []
        self._ldap_metrics = {}

        self.reconfigure()

    def reconfigure(self):
        self._configuration_dict = read_yaml_file(self._config_file_name)
        normalized_configuration = ConfigurationTransformationChainSingleton().transform_configuration(
            self._configuration_dict
        )

        self._sleep_time = normalized_configuration.get('period', 5)
        log_config = normalized_configuration.get('log_config')
        if log_config and isinstance(log_config, dict):
            log_config['version'] = log_config.get('version', 1)
            logging.config.dictConfig(log_config)
        for exporter_config in normalized_configuration.get('exporters', []):
            exporter = create_exporter(exporter_config)
            stats.stats.view_manager.register_exporter(exporter)

        for ldap_server_config in normalized_configuration.get('ldap_servers', []):
            if not ldap_server_config.get('sync_only', False):
                metric_set = self.generate_metric_set(ldap_server_config)
                self._metric_sets.append(metric_set)

        for base_dn, ldap_server_names in normalized_configuration.get('sync', {}).items():
            ldap_servers = [
                LdapServerPool().get_ldap_server(**server)
                for server in normalized_configuration.get('ldap_servers', [])
                if server['database'] in ldap_server_names
            ]
            sync_metric_set = SyncMetricSet(
                base_dn=base_dn,
                ldap_servers=ldap_servers
            )
            self._metric_sets.append(sync_metric_set)

    def generate_metric_set(self, ldap_server_config):
        args = copy.deepcopy(ldap_server_config.get('connection', {}))
        args['database'] = ldap_server_config.get('database')
        ldap_server = LdapServerPool().get_ldap_server(**args)
        metric_set = MetricSet(ldap_server=ldap_server)
        configs = [ldap_server_config.get('object')]
        while configs:
            config = configs.pop()
            if isinstance(config, list):
                for item in config:
                    configs.insert(0, item)
            elif isinstance(config, dict):
                dn = config.get('computed_dn')
                name = config.get('metric_name')
                attribute = config.get('attribute')
                unit = config.get('unit')
                description = config.get('description', '')
                value_function = config.get('func', 'value')
                query_dn = config.get('query_dn')
                if dn and name and attribute and unit and query_dn:
                    stat = self._ldap_metrics.get(name)
                    if not stat:
                        stat = LdapStatistic(
                            dn=dn,
                            name=name,
                            attribute=attribute,
                            description=description,
                            unit=unit,
                            value_function=value_function,
                            query_dn=query_dn,
                            tag_keys=['database']
                        )
                        self._ldap_metrics[name] = stat
                    metric_set.add_statistic(stat)
                for key, value in config.items():
                    configs.insert(0, value)
        return metric_set

    def metric_sets(self):
        return self._metric_sets

    def sleep(self):
        sleep(self._sleep_time)


def read_yaml_file(file_name):
    with open(file_name, 'r') as file:
        ret_val = yaml.safe_load(file)
    return ret_val


def create_exporter(exporter_configuration=None):
    if exporter_configuration is None:
        raise ValueError("Cannot create an exporter with no configuration!")

    name = exporter_configuration.get('name')
    if name not in SUPPORTED_EXPORTERS:
        logging.error(
            f"Requested exporter named {name}, which is not supported.  Choose from:{', '.join(SUPPORTED_EXPORTERS)}"
        )
        raise ValueError(
            f"Requested exporter named {name}, which is not supported.  Choose from:{', '.join(SUPPORTED_EXPORTERS)}"
        )

    options = exporter_configuration.get('options', {})
    exporter = None
    if "Prometheus" == name:
        if 'options' not in exporter_configuration:
            logging.error("The Prometheus exporter requires options configuration.")
            raise ValueError("The Prometheus exporter requires options configuration.")
        final_options = {'namespace': 'openldap', 'port': 8000, 'address': '0.0.0.0'}
        final_options.update(options)
        exporter = stats_exporter.new_stats_exporter(
            stats_exporter.Options(**final_options)
        )

    elif "Stackdriver" == name:
        exporter = opencensus.ext.stackdriver.stats_exporter.new_stats_exporter(interval=5)
        print(f"Exporting stats to this project {exporter.options.project_id}")

    return exporter



import copy

from opencensus.stats import stats
from opencensus.tags import tag_map, tag_value, tag_key


class MetricSet:
    def __init__(self, ldap_server=None, ldap_statistics=None):
        self._ldap_server = ldap_server
        if not ldap_statistics or not isinstance(ldap_statistics, list):
            ldap_statistics = []
        self._ldap_statistics = copy.deepcopy(ldap_statistics)
        self._query_dns = set()

    def set_ldap_server(self, ldap_server):
        self._ldap_server = ldap_server

    def add_statistic(self, ldap_statistic):
        self._ldap_statistics.append(ldap_statistic)
        self._query_dns.add(ldap_statistic.query_dn)

    def collect(self):
        tag_keys = [tag_key.TagKey('database')]
        results = {}
        for query_dn in self._query_dns:
            results[query_dn] = dict([
                (result_dn, result_attributes)
                for result_dn, result_attributes in self._ldap_server.query(dn=query_dn)
            ])
        mmap = stats.stats.stats_recorder.new_measurement_map()
        for server_statistic in self._ldap_statistics:
            ldap_value = results.get(
                server_statistic.query_dn, {}
            ).get(
                server_statistic.dn, {}
            ).get(
                server_statistic.attribute
            )
            server_statistic.collect(ldap_server=self._ldap_server, measurement_map=mmap, ldap_value=ldap_value)
        tmap = tag_map.TagMap()
        tmap.insert(
            tag_keys[0],
            tag_value.TagValue(self._ldap_server.database)
        )
        mmap.record(tmap)



import logging

import ldap


class LdapServerPool:
    _ldap_servers = {}

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(LdapServerPool, cls).__new__(cls)
        return cls.instance

    def get_ldap_server(self, **kwargs):
        if not self._ldap_servers.get(kwargs['database']):
            self._ldap_servers[kwargs['database']] = LdapServer(**kwargs)
            logging.critical(f"Registered LDAP Server: {kwargs['database']}")
        return self._ldap_servers[kwargs['database']]


class LdapServer:
    def __init__(self,
                 server_uri,
                 user_dn=None,
                 user_password=None,
                 database=None,
                 start_tls=False,
                 ca_file=None,
                 cert_file=None,
                 key_file=None,
                 sasl_mech=None,
                 timeout=-1):
        if database is None:
            database = server_uri

        self.connection = None
        self.database = database
        self.user_dn = user_dn
        self.user_password = user_password
        self.sasl_mech = sasl_mech

        self.connect_to_ldap(
            server_uri=server_uri,
            start_tls=start_tls,
            ca_file=ca_file,
            cert_file=cert_file,
            key_file=key_file,
            timeout=timeout
        )

    def connect_to_ldap(self, server_uri, start_tls=False, ca_file=None, cert_file=None, key_file=None, timeout=-1):
        if server_uri is None:
            logging.error(f"Failing to configure LDAP server {self.database} because no URI was supplied.")
            raise ValueError(f"An LDAP server URI must be defined for {self.database}")

        self.connection = ldap.ldapobject.ReconnectLDAPObject(server_uri)
        self.connection.timeout = timeout

        if ca_file:
            self.connection.set_option(ldap.OPT_X_TLS_CACERTFILE, ca_file)
        if cert_file:
            if not key_file:
                logging.error(f"Certificate file specified, but no key file specified for {self.database}")
                raise ValueError(f"Certificate file specified, but no key file specified for {self.database}")
            self.connection.set_option(ldap.OPT_X_TLS_CERTFILE, cert_file)
            self.connection.set_option(ldap.OPT_X_TLS_KEYFILE, key_file)
        if ca_file or cert_file:
            self.connection.set_option(ldap.OPT_X_TLS_NEWCTX, 0)

        if start_tls:
            logging.info(f"Using StartTLS for {self.database}")
            self.connection.protocol_version = ldap.VERSION3
            self.connection.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
            self.connection.start_tls_s()

    def query(self, dn=None, scope=ldap.SCOPE_SUBTREE, attr_list=None):
        logging.debug(f"Querying {self.database} for {dn}")
        if attr_list is None:
            attr_list = ['+']
        if dn is None:
            logging.error("INTERNAL ERROR: Could not run a query because no DN was supplied")
            raise ValueError('Must specify a DN to query')

        try:
            if self.sasl_mech:
                if self.sasl_mech == 'EXTERNAL':
                    self.connection.sasl_external_bind_s()
                else:
                    logging.error(f"INTERNAL ERROR: Unsupported SASL mechanism {self.sasl_mech}")
                    raise ValueError(f"Unsupported SASL mechanism {self.sasl_mech}")
            else:
                self.connection.simple_bind_s(self.user_dn, self.user_password)
            return self.connection.search_s(dn, scope=scope, attrlist=attr_list)
        except (ldap.SERVER_DOWN, ldap.NO_SUCH_OBJECT, ldap.TIMEOUT) as error:
            logging.error('Could not query LDAP:')
            logging.exception(error)
            return []

    def query_dn_and_attribute(self, dn, attribute):
        results = self.query(dn, scope=ldap.SCOPE_BASE, attr_list=[attribute])
        if not results:
            return None

        result_dn, result_attributes = results[0]
        if attribute not in result_attributes:
            return None
        return result_attributes.get(attribute)



import re

from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class SnakeCaseConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        config = {}
        for key, value in configuration.items():
            snake_case_key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
            if isinstance(value, dict):
                snake_cased_value = SnakeCaseConfigurationTransformer.process(value)
            elif isinstance(value, list):
                snake_cased_value = [
                    SnakeCaseConfigurationTransformer.process(x) if isinstance(x, dict) else x
                    for x in value
                ]
            else:
                snake_cased_value = value
            config[snake_case_key] = snake_cased_value
        return config



from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class MetricQueryDnConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        return MetricQueryDnConfigurationTransformer.process_object_config(configuration)

    @staticmethod
    def process_object_config(configuration, query_dn=None):
        if isinstance(configuration, dict):
            config = {}
            configuration_query_dn = configuration.get('query_dn', configuration.get('rdn'))
            if not query_dn and configuration_query_dn:
                query_dn = configuration_query_dn

            for key, value in configuration.items():
                config[key] = MetricQueryDnConfigurationTransformer.process_object_config(
                    configuration=value,
                    query_dn=query_dn,
                )

            # True if either 'rdn' or 'attribute' is a key
            if query_dn:
                config['query_dn'] = query_dn

            return config
        elif isinstance(configuration, list):
            return [
                MetricQueryDnConfigurationTransformer.process_object_config(
                    configuration=item,
                    query_dn=query_dn,
                )
                for item in configuration
            ]
        else:
            return configuration



import copy
import logging
import re

from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class MetricNameInterpolationConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        if not isinstance(configuration, dict):
            raise ValueError(
                'NameInterpolationConfigurationTransformer expect a configuration that is a dict '
                'but received one that isn\'t.')

        config = copy.deepcopy(configuration)
        for server_config in config.get('ldap_servers'):
            if not server_config.get('sync_only', False):
                ldap_server = ConfigurationTransformer.get_ldap_server(server_config=server_config)
                server_config['object'] = MetricNameInterpolationConfigurationTransformer.process_objects_for_ldap_server(
                    configuration=server_config.get('object', {}),
                    dn='',
                    ldap_server=ldap_server
                )
        return config

    @staticmethod
    def process_objects_for_ldap_server(configuration, dn, ldap_server):

        if isinstance(configuration, list):
            return [
                MetricNameInterpolationConfigurationTransformer.process_objects_for_ldap_server(item, dn, ldap_server)
                for item in configuration
            ]
        elif isinstance(configuration, dict):
            config = {}

            for key, value in configuration.items():
                proper_value = MetricNameInterpolationConfigurationTransformer.process_objects_for_ldap_server(
                    value, dn, ldap_server)
                if key in ['name']:
                    config['computed_name'] = MetricNameInterpolationConfigurationTransformer.create_metric_name(
                        configuration,
                        ldap_server
                    )
                config[key] = proper_value

            return config

        else:
            return configuration

    @staticmethod
    def create_metric_name(config, ldap_server):
        def repl_func(matches):
            groups = matches.groups()
            if 'rdn' == groups[0]:
                index = int(groups[1])
                rdn_matches = re.match(
                    config.get('rdn'),            # Pattern to match - the 'rdn' entry
                    config.get('computed_rdn')    # String for matching - the 'computed_rdn' entry
                )
                if rdn_matches:
                    return rdn_matches.group(index)
            elif 'attr' == groups[0]:
                attribute_name = groups[1]
                ldap_result = ldap_server.query_dn_and_attribute(
                    config.get('computed_dn', ''),
                    attribute_name
                )
                value = ldap_result[0]
                if isinstance(value, bytes):
                    value = value.decode()
                return str(value)
            else:
                logging.error(f"Unknown interpolation requested in name: {groups[0]}.{groups[1]}")

        name = config.get('computed_name', config.get('name', ''))
        name = re.sub(
            '{([^\\.}]+)\\.([^}]+)}',
            repl_func,
            name
        ).lower()
        return name



from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class MetricNameConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        return MetricNameConfigurationTransformer.process_object_config(configuration)

    @staticmethod
    def process_object_config(configuration, prefix='', default_name='', separator='/'):
        full_prefix = prefix + separator if prefix else ''
        if isinstance(configuration, dict):
            config = {}
            metric_name = prefix
            if configuration.get('rdn') or configuration.get('attribute'):
                metric_name = full_prefix + configuration.get('computed_name', configuration.get('name', default_name))
                config['metric_name'] = metric_name
            for key, value in configuration.items():
                config[key] = MetricNameConfigurationTransformer.process_object_config(
                    configuration=value,
                    prefix=metric_name,
                    default_name=key,
                    separator=separator
                )
            return config
        elif isinstance(configuration, list):
            return [
                MetricNameConfigurationTransformer.process_object_config(
                    configuration=item,
                    prefix=prefix,
                    default_name=default_name,
                    separator=separator
                )
                for item in configuration
            ]
        else:
            return configuration



# openldap_opencensus_stats
# Copyright (C) 2023  NetworkRADIUS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from opencensus.stats import measure, view, aggregation, stats


class LdapSyncStatistic:

    @staticmethod
    def log_and_raise(message=''):
        logging.error(message)
        raise ValueError(message)

    def __init__(self,
                 name=None,
                 description='Undescribed',
                 unit='ms',
                 tag_keys=None):
        if name is None:
            self.log_and_raise('Statistics definition must include a name for the statistic')
        self.name = name
        tag_keys = tag_keys or ['BaseDN']

        self.measure = self.generate_measure(description, name, unit)
        self.view = self.generate_view(description, name, tag_keys, self.measure)
        stats.stats.view_manager.register_view(self.view)

    @staticmethod
    def generate_view(description, name, tag_keys, stat_measure):
        return view.View(
            name=name,
            description=description,
            columns=tag_keys,
            aggregation=aggregation.LastValueAggregation(),
            measure=stat_measure
        )

    @staticmethod
    def generate_measure(description, name, unit):
        return measure.MeasureFloat(
            name=name,
            description=description,
            unit=unit
        )

    def display_name(self,
                     ldap_server=None):
        return self.name

    def collect(self,
                ldap_server=None,
                measurement_map=None,
                offset=None):
        if offset is None:
            logging.warning(f"No offset collected for {self.display_name(ldap_server)}")
            return
        if measurement_map is None:
            logging.warning(f"No measurement_map for {self.display_name(ldap_server)}")
            return

        logging.debug(f"Collected ldap_value for {self.display_name(ldap_server)}: {offset}")
        measurement_map.measure_float_put(self.measure, offset)



#!/usr/bin/python3

# openldap-opencensus-stats.py
# Copyright (C) 2023  NetworkRADIUS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
from openldap_opencensus_stats.configuration import Configuration


def parse_command_line():
    parser = argparse.ArgumentParser(description='Monitor the LDAP database.')
    parser.add_argument('config_file')

    return parser.parse_args()


def monitor():
    args = parse_command_line()
    configuration = Configuration(args.config_file)
    while True:
        metric_sets = configuration.metric_sets()
        for metric_set in metric_sets:
            metric_set.collect()
        configuration.sleep()


if __name__ == '__main__':
    monitor()



# openldap_opencensus_stats
# Copyright (C) 2023  NetworkRADIUS
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from opencensus.stats import measure, view, aggregation, stats


class LdapStatistic:

    @staticmethod
    def log_and_raise(message=''):
        logging.error(message)
        raise ValueError(message)

    def __init__(self,
                 dn=None,
                 name=None,
                 attribute=None,
                 description='Unspecified',
                 unit='By',
                 value_function='value',
                 query_dn=None,
                 tag_keys=None):
        if tag_keys is None:
            tag_keys = []
        if dn is None:
            self.log_and_raise('Statistics definition must include the dn attribute')
        if name is None:
            self.log_and_raise('Statistics definition must include a name for the statistic')
        if attribute is None:
            self.log_and_raise('Statistics definition must include the attribute to query')
        if query_dn is None:
            self.log_and_raise('Statistics definition must include the DN to query')

        self.attribute = attribute
        self.dn = dn
        self.query_dn = query_dn
        self.measure = measure.MeasureFloat(
            name=name,
            description=description,
            unit=unit
        )

        self.view = view.View(
            name=name,
            description=description,
            columns=tag_keys,
            aggregation=aggregation.LastValueAggregation(),
            measure=self.measure
        )
        stats.stats.view_manager.register_view(self.view)
        self._value_function = lambda value: eval(value_function)

    def display_name(self):
        return f"{self.measure.name}:{self.attribute}"

    def collect(self, ldap_server=None, measurement_map=None, ldap_value=None):
        def display_name(server, statistic):
            return f"{server.database}:{statistic.display_name()}"

        if ldap_server is None:
            self.log_and_raise(f"INTERNAL ERROR: Failing to collect statistic {self.display_name()} "
                               f"because no LDAP server supplied.")
        if measurement_map is None:
            self.log_and_raise(f"INTERNAL ERROR: Failing to collect statistic {self.display_name()} "
                               f"because no measurement map was supplied.")

        # if not ldap_value:
        #     ldap_value = ldap_server.query_dn_and_attribute(self.dn, self.attribute)
        if ldap_value is None:
            logging.warning(f"No ldap_value collected for {display_name(ldap_server, self)}")
            return
        ldap_value_float = float(ldap_value[0])
        logging.debug(f"Collected ldap_value for {display_name(ldap_server, self)}: {ldap_value_float}")
        value = self._value_function(ldap_value_float)
        if ldap_value_float != value:
            logging.debug(f"  Transformed into: {value}")
        measurement_map.measure_float_put(self.measure, value)



from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class MetricDnConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        return MetricDnConfigurationTransformer.process_object_config(configuration)

    @staticmethod
    def process_object_config(configuration, suffix_dn='', separator=','):
        if isinstance(configuration, dict):
            config = {}
            if suffix_dn:
                full_suffix = separator + suffix_dn
            else:
                full_suffix = suffix_dn
            computed_dn = suffix_dn
            rdn = configuration.get('computed_rdn', configuration.get('rdn'))
            if rdn:
                computed_dn = rdn + full_suffix

            for key, value in configuration.items():
                config[key] = MetricDnConfigurationTransformer.process_object_config(
                    configuration=value,
                    suffix_dn=computed_dn,
                    separator=separator
                )

            # True if either 'rdn' or 'attribute' is a key
            if configuration.get('rdn', configuration.get('attribute')):
                config['computed_dn'] = computed_dn
            else:
                computed_dn = suffix_dn

            return config
        elif isinstance(configuration, list):
            return [
                MetricDnConfigurationTransformer.process_object_config(
                    configuration=item,
                    suffix_dn=suffix_dn,
                    separator=separator
                )
                for item in configuration
            ]
        else:
            return configuration



import copy

from openldap_opencensus_stats.config_transformers.base import ConfigurationTransformer


class CopyForEachDatabaseConfigurationTransformer(ConfigurationTransformer):
    @staticmethod
    def process(configuration):
        conf = copy.deepcopy(configuration)
        ldap_servers = conf.get('ldap_servers', [])
        for server_conf in ldap_servers:
            server_conf['object'] = copy.deepcopy(conf.get('object'))
        return conf



from setuptools import setup, find_packages

VERSION = '0.0.10'
DESCRIPTION = 'OpenLDAP OpenCensus Statistics'
LONG_DESCRIPTION = 'A package to gather statistics for OpenLDAP and publish via OpenCensus.'


setup(
    name="openldap-opencensus-stats",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Mark Donnelly",
    author_email="mark@painless-security.com",
    license='AGPL 3.0',
    packages=find_packages(),
    install_requires=[
        'grpcio==1.47.0',
        'opencensus-ext-stackdriver==0.8.0',
        'opencensus-ext-prometheus',
        'opencensus==0.10.0',
        'python-ldap',
        'pyyaml',
    ],
    keywords='openldap opencensus metrics',
    entry_points={
        'console_scripts': [
            'openldap_opencensus_stats = openldap_opencensus_stats.main:monitor'
        ]
    },
    data_files=[
        ('etc', ['openldap-opencensus-stats.template.yml']),
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: No Input/Output (Daemon)",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        "Programming Language :: Python :: 3",
        "Topic :: System :: Monitoring",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    ]
)
